library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

--v2 added overscan support
--v3 (24/07/2000) hclock reduced to 500kHz, vclock reduced to 125kHz

entity cpix is
    port (
        oscin: in STD_LOGIC;		-- 12 MHz oscillator
        cpumpck: out STD_LOGIC;		-- 1Mhz out for charge pump
        
        sdat: out STD_LOGIC;		-- serial data to the ADC
        sclk: out STD_LOGIC;		-- serial data to the ADC
        sload: out STD_LOGIC;		-- serial data to the ADC

        adck: out STD_LOGIC;		-- ADC sample clock
        cdsck1: out STD_LOGIC;		-- ADC CDS clock
        cdsck2: out STD_LOGIC;		-- ADC CDS clock
        
        sag: out STD_LOGIC;		-- 1'st ccd clocks
        iag1: out STD_LOGIC;		-- 1'st ccd clocks
        iag2: out STD_LOGIC;		-- 1'st ccd clocks
        rst: out STD_LOGIC;		-- 1'st ccd clocks
        odb: out STD_LOGIC;		-- 1'st ccd clocks
        srg: out STD_LOGIC;		-- 1'st ccd clocks

        sda: inout STD_LOGIC;		-- I2c to temp controller
        scl: out STD_LOGIC;		-- I2c to temp controller

	rwi: in STD_LOGIC;		-- parallel port strobes
	datast: in STD_LOGIC;		-- parallel port strobes
	addrst: in STD_LOGIC;		-- parallel port strobes
	waitn: out STD_LOGIC;		-- parallel port strobes

	io: inout STD_LOGIC_VECTOR(7 downto 0);	-- parallel port data
	
	tcr: in STD_LOGIC;		-- crit temp input.
	
	fd: in STD_LOGIC_VECTOR(7 downto 0);	-- data from fifo
	
	fef: in STD_LOGIC;		-- fifo empty flag.
	fpae: in STD_LOGIC;		-- fifo partial empty flag.
	foe: out STD_LOGIC;		-- fifo output enable.
	frck: out STD_LOGIC;		-- fifo read clock.
	fren: out STD_LOGIC;		-- fifo read enable.
	
	fwck: out STD_LOGIC;		-- fifo write clock.
	fwen: out STD_LOGIC;		-- fifo write enable.
	frs: out STD_LOGIC;		-- fifo reset.

	ckout: out STD_LOGIC;		-- clock/2
	usbrd: inout STD_LOGIC;		-- usb fifo controls
	usbwr: inout STD_LOGIC;		-- 
	usbren: in STD_LOGIC;		-- 
	usbwen: in STD_LOGIC		-- 
	
    );
end cpix;

architecture cpix_arch of cpix is

type ccdst_type is (integrate, vsh1, vsh2, vse, vsel1, vsel2, 
vsel3, vsel4, hs1, hs2, fwa);

type eppst_type is (idle, sfd, sfa, rd, wr, rend, wend);

signal eppst: eppst_type;


signal oscdiv: STD_LOGIC;			-- oscilator divider

signal clkdiv: STD_LOGIC_VECTOR(5 downto 0); 	-- clock divider
signal addreg: STD_LOGIC_VECTOR(3 downto 0);	-- parallel port address register
signal linecnt: STD_LOGIC_VECTOR(9 downto 0);	-- ccd line counter
signal colcnt: STD_LOGIC_VECTOR(9 downto 0);	-- ccd column counter
signal reset: STD_LOGIC;			-- general software reset
signal linelimit: STD_LOGIC_VECTOR(7 downto 0); -- (500 + first line to read) / 4
signal readst: STD_LOGIC;			-- ccd read start signal
signal ccdst: ccdst_type;
signal resen: STD_LOGIC;			-- flag to enable reset
signal ck2: STD_LOGIC;				-- 2MHz clock

signal ckdiv1: STD_LOGIC;			
signal ckdiv2: STD_LOGIC;			
signal fweni: STD_LOGIC;
--signal datasti: STD_LOGIC;
signal datastd: STD_LOGIC;

--signal addrsti: STD_LOGIC;
signal addrstd: STD_LOGIC;

signal wrstrobe : std_logic;
signal awrstrobe : std_logic;
signal rstrobe : std_logic;

signal sdaout: STD_LOGIC;
signal newline: STD_LOGIC;

signal dataav: STD_LOGIC;
signal fdiv: STD_LOGIC_VECTOR(1 downto 0);

signal pwres: std_logic;

--signal write: STD_LOGIC;
--signal rwid: std_logic;

signal overscan: STD_LOGIC;	-- enable reading of an extra 64 pixels per line when 1

signal testreg: std_logic_vector(7 downto 0);
signal ind: std_logic_vector(7 downto 0);
signal din: std_logic_vector(7 downto 0);

begin
  -- <<enter your statements here>>

fwck <= not clkdiv(2);
rst <= '0' when resen = '1' and clkdiv(3 downto 0) = "1000" else '1';
cdsck1 <= '1' when resen = '1' and clkdiv(3 downto 0) = "1101" else '0';

ck2 <= not clkdiv(2);

foe <= '0';

frck <= not oscin;
	

io <= fd when addreg = "0000" and datast = '0' and rwi = '1' else -- read data from fifo
      "0000" & fef & fpae & sda & "0" when addreg = "0011" and datast = '0' and rwi = '1' else
      not testreg when addreg = "0100" and datast = '0' and rwi = '1' else
	"ZZZZZZZZ";

sda <= '0' when sdaout = '0' else 'Z';

--addrsti <= '0' when addrstd = '1' and addrstdd = '0' else '1';
--datasti <= '0' when datastd = '1' and datastdd = '0' else '1';

fren <='0' when addreg = "0000" and rstrobe = '1'
    else '1';   

ckout <= oscdiv;

process (oscin)	-- address register
begin
    if rising_edge(oscin) then
	if awrstrobe = '1' then
		addreg(3 downto 0) <= io(3 downto 0);
	end if;
	end if;
end process;

process (oscin)	-- register writes 
begin
    if rising_edge(oscin) then
	if wrstrobe = '1' and addreg = "0000" then
		sload <= io(2);		-- adc config (addr: 0)
		sclk <= io(1);
		sdat <= io(0);
	elsif wrstrobe = '1' and addreg = "0001" then
		overscan <= io(4);
		newline <= io(3);
		odb <= io(2);		-- reset and readout start (adr: 1)
		readst <= io(1);
		reset <= io(0);
	elsif wrstrobe = '1' and addreg = "0010" then
		linelimit <= io;	-- start of readout area (adr: 2)
	elsif wrstrobe = '1' and addreg = "0011" then
		scl <= io(0);	-- i2c for temperature readout (adr: 3)
		sdaout <= io(1);
	elsif wrstrobe = '1' and addreg = "0100" then
		testreg <= io;	-- testreg (adr: 4)
	end if;
    end if;
end process;

process (oscin)
begin
	if rising_edge(oscin) then
		oscdiv <= not oscdiv;
	end if;
end process;

process(oscdiv)
begin
    if rising_edge(oscdiv) then
	fdiv<= fdiv + 1;
	if (fdiv = 0) then
	    dataav <= '1';
	end if;
	if dataav = '1' and usbwen = '0' then
	    usbwr <= '1';
	    dataav <= '0';
	end if;
	if (usbwr = '1') then
	    usbwr <= '0';
	end if;
	if usbren = '0' then
	    usbrd <= '1';
	end if;
	if usbrd = '1' then
	    usbrd <= '0';
	end if;
    end if;
end process;


process (oscdiv, clkdiv, reset)	-- readout state machine
begin
    if (reset = '1') then
    	iag1 <= '1' ;
    	iag2 <= '1' ;
    	sag <= '1' ;
    	srg <= '1' ;
    	ccdst <= integrate ;
    elsif (rising_edge(oscdiv)) then
    	if clkdiv(2 downto 0) = "110" then
	case ccdst is
		when integrate =>
			fweni <= '1';
			resen <= '0';
			iag1 <= '1';
			iag2 <= '1';
			sag <= '1';
			srg <= '1';
			linecnt <= "1111101000";
			if readst = '1' and clkdiv(5 downto 3) = "001" then 
				ccdst <= vsh1;
				frs <= '0';
			end if;
		when vsh1 =>
			frs <= '1';
			resen <= '1';
			iag1 <= '0';
			iag2 <= '0';
			sag <= '1';
			srg <= '0';
			if clkdiv(5 downto 3) = "001" then 
				linecnt <= linecnt - 1;
				ccdst <= vsh2;
			end if;
		when vsh2 =>
			iag1 <= '1';
			iag2 <= '1';
			sag <= '0';
			srg <= '1';
			if clkdiv(5 downto 3) = "001" then 
				if linecnt(9 downto 2) = linelimit then ccdst <= vse;
				else ccdst <= vsh1;
				end if;
			end if;
		when vse =>
			sag <= '1';
			if clkdiv(5 downto 3) = "001" then 
				ccdst <= vsel1;
			end if;
		when vsel1 =>
			frs <= '0';
			sag <= '0';
			if clkdiv(5 downto 3) = "001" then 
				linecnt <= linecnt - 1;
				colcnt <=  "101" & overscan & "101100";
				ccdst <= vsel2;
			end if;
		when vsel2 =>
			frs <= '1';
			srg <= '0';
			if clkdiv(5 downto 3) = "001" then 
				ccdst <= vsel3;	
			end if;
		when vsel3 =>
			sag <= '1';
			if clkdiv(5 downto 3) = "001" then 
				ccdst <= vsel4;
			end if;
		when vsel4 =>
			srg <= '1';
			if clkdiv(5 downto 3) = "001" then ccdst <= hs1;
			end if;
		when hs1 =>
			srg <= '0';
			colcnt <= colcnt - 1;
			ccdst <= hs2;
			fweni <= '0';
		when hs2 =>
			srg <= '1';
			if colcnt = "0000000000" then
				if linecnt = 0 then
				    ccdst <= integrate;
				else
				    ccdst <= fwa;
				end if;
			else
			    ccdst <= hs1;
			end if;
		when fwa =>
			fweni <= '1';
			if fef = '0' and newline = '1' and clkdiv(5 downto 3) = "000" then
			    ccdst <= vsel1;
			end if;
		end case;
		end if;
    end if;
end process;
  
process (oscdiv)			-- clock divider & pixel logic
begin
    if rising_edge(oscdiv) then
	cpumpck <= clkdiv(3);
	if clkdiv(2 downto 0) = "010" then
		clkdiv(2 downto 0) <= "100";
	elsif clkdiv(2 downto 0) = "110" then
		clkdiv(2 downto 0) <= "000";
		clkdiv(5 downto 3) <= clkdiv(5 downto 3) + 1;
	else
	    clkdiv(2 downto 0) <= clkdiv(2 downto 0) + 1;
	end if;
	ckdiv1 <= not ckdiv1;
	ckdiv2 <= not ckdiv2;
	if clkdiv(3 downto 0) = "0010" then
	    adck <= '1';
	end if;
	if clkdiv(3 downto 0) = "1010" then
	    adck <= '0';
	end if;
	if clkdiv(3 downto 0) = "0001" then
	    cdsck2 <= '1';
	end if;
	if clkdiv(3 downto 0) = "0101" then
	    cdsck2 <= '0';
	end if;
	if clkdiv(3 downto 0) = "0010" then
		if fweni = '1' then
		    fwen <= '1';
		else
		    fwen <= '0';
		end if;
	end if;
     end if;
end process;

-- epp interface logic


process (oscin)	-- epp state machine
begin
    if (rising_edge(oscin)) then
      if (pwres = '0') then
    	eppst <= idle ;
       wrstrobe <= '0';
      rstrobe <= '0';
     awrstrobe <= '0';
    waitn <= '0';
      else
        case eppst is
          when idle =>
            wrstrobe <= '0';
            rstrobe <= '0';
            awrstrobe <= '0';
            waitn <= '0';
            if datastd = '0' then
              eppst <= sfd;
            elsif addrstd = '0' then
              eppst <= sfa;
            end if;
          when sfd =>                   -- data strobe filter 
            if datastd = '0' then
              if rwi = '0' then
                eppst <= wr;
                wrstrobe <= '1';
              else
                eppst <= rd;
              end if;
            else
              eppst <= idle;          
            end if;
          when sfa =>                   -- adress strobe filter
            if addrstd = '0' then
              if rwi = '0' then
                eppst <= wr;
                awrstrobe <= '1';
              else
                eppst <= idle;
              end if;
            else
              eppst <= idle;          
            end if;
          when wr =>                    -- write
            waitn <= '1';
            eppst <= wend;
            wrstrobe <= '0';
            awrstrobe <= '0';
          when wend =>                 -- write end
            if datastd = '1' and addrstd = '1' then
              eppst <= idle;
              waitn <= '0';
            end if;
          when rd =>
            waitn <= '1';
            eppst <= rend;              -- read
          when rend =>                 -- write end
            if datastd = '1' then
              waitn <= '0';
              eppst <= idle;
              rstrobe <= '1';
            end if;
			when others =>
				eppst <= idle;
        end case;
    end if;
	end if;
end process;        


process (oscin)
begin
    if rising_edge(oscin) then
	datastd <= datast;
	addrstd <= addrst;
	if (addrstd = '1') then
		pwres <= '1';
	elsif (datastd = '0') then
		pwres <= '0';
	end if;
    end if;
end process;

end cpix_arch;

